import re


class S3NamingHelper:

    def validate_part(self, value, allow_prefix: bool = True) -> tuple:
        '''checks if a valid s3 part
            ARGS:
                - value (str) the path or partial path in s3
                - allow_prefix (bool) if false will only validate single parts
            RETURNS: tuple validation and reason value
        '''
        if not allow_prefix and '/' in value:
            return tuple([False, 'prefix dissalowed'])
        if len(value) < 1:
            return tuple([False, 'path parts must be 1 or more characters'])

        vals = value.split('/')
        for part in vals:
            # get rid of those leading / trailing slashes
            if len(part) > 0:
                for c in part:
                    if c not in self._safe_chars():
                        return tuple([False, f'{c} is not an allowed character.'])

        return tuple([True, None])

    def validate_bucket_name(self, bucket_name: str):
        result, message = self._validate_bucket_name(bucket_name)
        if result:
            return result
        else:
            raise ValueError(message)

    def _validate_bucket_name(self, bucket_name: str) -> tuple:
        ''' INTENT: checks for a valid bucket name
            ARGS: 
                - bucket_name (str) the bucket name to validate
            RETURNS: tuple validation and reason value
        '''
        # must be between 3-63 chars
        if len(bucket_name) < 3 or len(bucket_name) > 63:
            return tuple([False, 'bucket name must be between 3 and 63 chars'])

        # lower case chars, numbers, periods, dashes
        elif not bucket_name.islower():
            return tuple([False, 'bucket name cannot contain upper case characters'])

        elif not bool(re.match(r"^[a-z0-9\-\.]*$", bucket_name)):
            return tuple([False, 'bucket name can only contain lower case chars, numbers, dashes and periods'])

        # cannot end with dash
        elif bucket_name.endswith('-'):
            return tuple([False, 'bucket name cannot end with a dash'])

        # cannot consecutive periods
        elif '..' in bucket_name:
            return (False, 'bucket name cannot include double periods')

        # dashes next to periods
        elif '.-' in bucket_name or '-.' in bucket_name:
            return (False, 'bucket name cannot have dashes next to periods')

        # char or number after period
        elif bool(re.search(r"\.[^0-9a-z]*", bucket_name)):
            return tuple([False, 'bucket name must have only a letter or a number after a period'])

        # char or number at start
        elif not (bucket_name[0].isalpha() or bucket_name[0].isnumeric()):
            return tuple([False, 'bucket name must start with a number or letter'])
        else:
            return tuple([True, None])

    def validate_s3_path(self, path: str) -> tuple:
        ''' INTENT: validate a complete s3 path
            ARGS:
                - path (str) the S3 path to validate
            RETURNS: tuple validation and reason value
        '''
        if path[:5] != 's3://':
            return tuple([False, 'bucket path must have arn prefix (s3://)'])

        path_parts = path[5:].split('/')

        bucket_validity = self._validate_bucket_name(path_parts[0])

        if not bucket_validity[0]:
            return tuple([False, bucket_validity[1]])

        for part in path_parts[1:]:
            part_validity = self.validate_part(part)
            if not part_validity[0]:
                return tuple([False, part_validity[1]])

        return tuple([True, path])

    def _safe_chars(self) -> list:
        safe = list(range(ord('a'), ord('z')+1))
        safe += list(range(ord('A'), ord('Z')+1))
        safe = [chr(x) for x in safe]
        safe += [str(x) for x in range(0, 10)]
        safe += ['!', '-', '_', '.', '*', '(', ')', '=']
        return safe



import boto3
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import s3fs
import sys
import logging
from typing import List


logger = logging.getLogger(__name__)


def check_dataframe_for_timedelta(dataframe: pd.DataFrame)->None:
    """Pyarrow can't do timedelta."""
    dtypes = set(dataframe.columns)
    for dtype in dtypes:
        if 'timedelta' in dtype:
            timedelta_message = "Pyarrow does not support parquet conversion of timedelta columns at this time."
            logger.debug(timedelta_message)
            raise NotImplementedError(timedelta_message)


def _check_partition_compatibility(partition: str) -> bool:
    """ Make sure each partition value is hive-allowed."""
    reserved = "ALL, ALTER, AND, ARRAY, AS, AUTHORIZATION, BETWEEN, BIGINT, BINARY, BOOLEAN, BOTH, BY, CASE, CAST, CHAR, COLUMN, CONF, CREATE, CROSS, CUBE, CURRENT, CURRENT_DATE, CURRENT_TIMESTAMP, CURSOR, DATABASE, DATE, DECIMAL, DELETE, DESCRIBE, DISTINCT, DOUBLE, DROP, ELSE, END, EXCHANGE, EXISTS, EXTENDED, EXTERNAL, FALSE, FETCH, FLOAT, FOLLOWING, FOR, FROM, FULL, FUNCTION, GRANT, GROUP, GROUPING, HAVING, IF, IMPORT, IN, INNER, INSERT, INT, INTERSECT, INTERVAL, INTO, IS, JOIN, LATERAL, LEFT, LESS, LIKE, LOCAL, MACRO, MAP, MORE, NONE, NOT, NULL, OF, ON, OR, ORDER, OUT, OUTER, OVER, PARTIALSCAN, PARTITION, PERCENT, PRECEDING, PRESERVE, PROCEDURE, RANGE, READS, REDUCE, REVOKE, RIGHT, ROLLUP, ROW, ROWS, SELECT, SET, SMALLINT, TABLE, TABLESAMPLE, THEN, TIMESTAMP, TO, TRANSFORM, TRIGGER, TRUE, TRUNCATE, UNBOUNDED, UNION, UNIQUEJOIN, UPDATE, USER, USING, UTC_TMESTAMP, VALUES, VARCHAR, WHEN, WHERE, WINDOW, WITH, COMMIT, ONLY, REGEXP, RLIKE, ROLLBACK, START, CACHE, CONSTRAINT, FOREIGN, PRIMARY, REFERENCES, DAYOFWEEK, EXTRACT, FLOOR, INTEGER, PRECISION, VIEWS, TIME, NUMERIC, SYNC".split()
    reserved = [x.strip(',') for x in reserved]
    return not (partition.upper() in reserved)


def check_partitions(partitions: iter, dataframe: pd.DataFrame)->None:
    logger.debug("Checking partition args...")
    for partition in partitions:
        if partition not in dataframe.columns.tolist():
            partition_message = f"Cannot set {partition} as a partition; this is not a valid column header for the supplied dataframe."
            logger.critical(partition_message)
            raise ValueError(partition_message)
        if not _check_partition_compatibility(partition):
            partition_message = f"{partition} is a reserved word in hive that cannot be used as a partition."
            logger.critical(partition_message)
            raise ValueError(partition_message)
    logger.debug("Done checking partitions.")


def s3_url(bucket: str, key: str):
    return '/'.join(["s3:/", bucket, key])


def _gen_parquet_to_s3(bucket: str, key: str, dataframe: pd.DataFrame,
                       partitions: list) -> None:
    """ pushes the parquet dataset directly to s3. """
    logger.info("Writing to S3...")
    table = pa.Table.from_pandas(dataframe, preserve_index=False)

    uri = s3_url(bucket, key)
    logger.debug(f"Writing to s3 location: {uri}...")
    pq.write_to_dataset(table, compression="snappy", root_path=uri,
                        partition_cols=partitions, filesystem=s3fs.S3FileSystem())
    logger.debug("Done writing to location.")


def _assign_partition_meta(bucket: str, key: str, dataframe: pd.DataFrame, partitions: iter) -> List[str]:
    """ assigns the dataset partition meta to all keys in the dataset"""
    s3_client = boto3.client('s3')
    all_files = []
    for obj in s3_client.list_objects(Bucket=bucket, Prefix=key)['Contents']:
        if obj['Key'].endswith(".parquet"):
            all_files.append(obj['Key'])

    for obj in all_files:
        logger.debug(f"Appending metadata to file {obj}..")
        s3_client.copy_object(Bucket=bucket, CopySource={'Bucket': bucket, 'Key': obj}, Key=obj,
                              Metadata={'partition_data_types': str(
                                  _parse_dataframe_col_types(
                                      dataframe=dataframe, partitions=partitions)
                              )}, MetadataDirective='REPLACE')
        logger.debug("Done appending metadata.")
    return all_files


def _parse_dataframe_col_types(dataframe: pd.DataFrame, partitions: iter) -> dict:
    """ Returns a dict with the column names as keys, the data types (in strings) as values."""
    logger.debug("Determining write metadata for publish...")
    dataframe = pd.DataFrame(dataframe[partitions])
    dtypes = {}
    for col, dtype in dataframe.dtypes.items():
        dtype = str(dtype)
        if dtype == 'object':
            dtypes[col] = 'string'
        elif dtype.startswith('int'):
            dtypes[col] = 'integer'
        elif dtype.startswith('float'):
            dtypes[col] = 'float'
        elif dtype.startswith('date'):
            dtypes[col] = 'datetime'
        elif dtype.startswith('category'):
            dtypes[col] = 'category'
        elif dtype == 'bool':
            dtypes[col] = 'boolean'
    logger.debug(f"Done.Metadata set as {dtypes}")
    return dtypes


def _sized_dataframes(dataframe: pd.DataFrame) -> tuple:
    """Takes a dataframe and slices it into sized dataframes for optimal parquet sizes in S3.
        RETURNS a tuple of dataframes.         
    """

    def log_size_estimate(num_bytes):
        if num_bytes == 0:
            return "0MB"
        elif round(num_bytes / float(1 << 20), 2) == 0.0:
            return "<0.1MB"
        else:
            return str(round(num_bytes / float(1 << 20), 2)) + "MB"

    # get first row size
    row_size_est = sys.getsizeof(dataframe.head(1))
    # get number of rows
    num_rows = int(dataframe.shape[0])
    frame_size_est = row_size_est * num_rows
    # at scale dataframes seem to compress around 3.5-4.5 times as parquet.
    # TODO: should build a real regression to calc this!
    compression_ratio = 4
    # 60MB compressed is ideal for Spectrum
    ideal_size = compression_ratio * (60 * float(1 << 20))
    # short circut if < ideal size
    batch_log_message = """row size estimate: {row_size_est} bytes.
number of rows: {num_rows} rows
frame size estimate: {frame_size_est} bytes
compression ratio: {compression_ratio}:1
ideal size: {ideal_size} bytes
"""
    logger.debug(batch_log_message)
    if ideal_size > frame_size_est:
        return tuple([dataframe])

    # math the number of estimated partitions
    sized_frames = []
    num_partitions = int(row_size_est * num_rows / ideal_size)
    rows_per_partition = int(num_rows / num_partitions)
    # for each partition do the thing
    for index, lower in enumerate(range(0, num_rows, rows_per_partition)):
        lower = lower if lower == 0 else lower + 1
        if index + 1 == num_partitions:
            upper = num_rows
        else:
            upper = lower + rows_per_partition
        sized_frames.append(dataframe[lower:upper])
    logger.info(f"sized out {len(sized_frames)} dataframes.")
    return tuple(sized_frames)


def publish(bucket: str, key: str, partitions: iter, dataframe: pd.DataFrame) -> None:
    logger.info("Checking params...")
    check_dataframe_for_timedelta(dataframe)
    check_partitions(partitions, dataframe)
    logger.info("Params valid.")
    logger.debug("Begin writing to S3..")

    files = []
    for frame in _sized_dataframes(dataframe):
        _gen_parquet_to_s3(bucket=bucket,
                           key=key,
                           dataframe=dataframe,
                           partitions=partitions)

        published_files = _assign_partition_meta(bucket=bucket,
                                                 key=key,
                                                 dataframe=dataframe,
                                                 partitions=partitions)
        files = files + published_files
    logger.debug("Done writing to S3.")
    return files



import ast
import boto3
from collections import OrderedDict
import datetime
import operator
from typing import Dict, List, Any
import multiprocessing as mp
from multiprocessing import get_context
import s3fs
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from .s3_naming_helper import S3NamingHelper

''' S3 Parquet to Dataframe Fetcher.
This class handles the portion of work that will return a concatenated 
dataframe based on the partition filters the specified dataset.

Required kwargs:
    bucket (str): S3 Bucket name
    key (str): S3 key that leads to the desired dataset
    filters (List(Filter)): filters to be applied to the Dataset Partitions
        Filters have the fields:
            partition (str):
                Partition to be applied to
            comparison (str):
                Comparison function - one of: [ == , != , > , < , >= , <= ]
            values (List(any)): 
                Values to compare to - must match partition data type
                May not use multiple values with the '<', '>' comparisons
'''


# Filter
# class Filter(NamedTuple):
#    partition: str
#    comparison: str
#    values: List[any]

Filter = {
    "partition": str,
    "comparison": str,
    "values": List[any]
}

OPS = {
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    "<=": operator.le,
    ">": operator.gt,
    "<": operator.lt
}

NON_NUM_TYPES = [
    "string",
    "category",
    "bool",
    "boolean"
]

''' TODO: development notes, remove after
Internal attributes:
    List of keys for the filtered dataset
    List of local files pulled from above
    List of dataframes

Phase 1:
    Take and validate input
    Get partitions
Phase 2:
    Compare partitions to filters
    Create file paths to filtered parquets
Phase 3:
    Pull down files
    Transform files to dataframes
    Concat dataframes and return
'''


def get_all_partition_values(bucket: str, key: str, partition: str) -> iter:
    """retruns all values, correctly typed, for a given partition IN NO ORDER."""
    all_files = _get_all_files_list(bucket, key)
    partition_dtype = _get_partitions_and_types(
        first_file_key=all_files[0], bucket=bucket)[partition]
    partition_values = _parse_partitions_and_values(all_files, key=key)[
        partition]
    return [convert_type(val, partition_dtype) for val in partition_values]


def get_diff_partition_values(bucket: str, key: str, partition: str, values_to_diff: iter, reverse: bool = False) -> iter:
    """ returns all the partition values in bucket/key that are not in values_to_diff.
        ARGS:
            values_to_diff: the iterable of values to compare the partition values to
            reverse: if True, will look for the values in values_to_diff that are not in partition values (basically backwards)    
     """
    all_files = _get_all_files_list(bucket, key)
    partition_dtype = _get_partitions_and_types(
        first_file_key=all_files[0], bucket=bucket)[partition]
    partition_values = _parse_partitions_and_values(all_files, key=key)[
        partition]

    partition_set = set(partition_values)
    values_to_diff_set = set([str(val) for val in values_to_diff])

    if reverse:
        diff = values_to_diff_set - partition_set
    else:
        diff = partition_set - values_to_diff_set
    return [convert_type(val, partition_dtype) for val in diff]


def get_max_partition_value(bucket: str, key: str, partition: str) -> any:
    ''' Returns the max value of the specified partition
    in the data type from the metadata.
    '''
    S3NamingHelper().validate_bucket_name(bucket_name=bucket)

    all_files = _get_all_files_list(bucket=bucket, key=key)

    partition_dtype = _get_partitions_and_types(
        first_file_key=all_files[0], bucket=bucket)[partition]
    partition_values = _parse_partitions_and_values(
        file_paths=all_files, key=key)[partition]

    if partition_dtype in NON_NUM_TYPES:
        raise ValueError(
            f"Max cannot be used on partition types of {partition_dtype}")

    return max([convert_type(val, partition_dtype) for val in partition_values])


def fetch(bucket: str, key: str, filters: List[type(Filter)] = {}, parallel: bool = True):
    ''' Access function to kick off all bits and return result. '''
    _validate_filter_rules(filters)
    S3NamingHelper().validate_bucket_name(bucket)

    all_files = _get_all_files_list(bucket, key)

    partition_metadata = _get_partitions_and_types(all_files[0], bucket)

    _validate_matching_filter_data_type(partition_metadata, filters)

    # strip out the filenames (which we need)
    partition_values = _parse_partitions_and_values(all_files, key)

    typed_values = _get_partition_value_data_types(
        partition_values, partition_metadata)

    # filtered_paths is a list of the S3 prefixes from which we want to load data
    filtered_paths = _get_filtered_key_list(typed_values, filters, key)

    files_to_load = []

    for file in all_files:
        for prefix in filtered_paths:
            if file.startswith(prefix):
                files_to_load.append(file)
                continue

    if not (files_to_load):
        empty_df_data = OrderedDict()

        for column, data in partition_metadata.items():
            data = dtype_to_pandas_dtype(data)
            empty_df_data.update({column: pd.Series([], dtype=data)})

        typed_empty_df = pd.DataFrame(empty_df_data)
        return typed_empty_df

    return _get_filtered_data(bucket=bucket, paths=files_to_load, partition_metadata=partition_metadata,
                              parallel=parallel)


def fetch_diff(input_bucket: str, input_key: str, comparison_bucket: str, comparison_key: str, partition: str, parallel: bool = True) -> pd.DataFrame:
    ''' Returns a dataframe of whats in the input dataset but not the comparison dataset by the specified partition
        ARGS:
            input_bucket (str): the bucket of the dataset to start from
            input_key (str): the key to the dataset to start from
            comparison_bucket (str): the bucket of the dataset to compare against
            comparison_key (str): the key to the dataset to compare against
            partition (str): the partition whos values to compare
    '''
    S3NamingHelper().validate_bucket_name(input_bucket)
    S3NamingHelper().validate_bucket_name(comparison_bucket)

    comparison_values = get_all_partition_values(
        bucket=comparison_bucket, key=comparison_key, partition=partition)

    diff_values = get_diff_partition_values(
        bucket=input_bucket,
        key=input_key,
        partition=partition,
        values_to_diff=comparison_values
    )

    filters = [{
        "partition": partition,
        "comparison": "==",
        "values": diff_values
    }]

    return fetch(bucket=input_bucket, key=input_key, filters=filters, parallel=parallel)


def convert_type(val: Any, dtype: str) -> Any:
    """ converts a value to the given datatype"""
    if dtype == 'string':
        return str(val)
    elif dtype == 'integer':
        return int(val)
    elif dtype == 'float':
        return float(val)
    elif dtype == 'datetime':
        return datetime.datetime.strptime(
            val, '%Y-%m-%d %H:%M:%S')
    elif dtype == 'category':
        return pd.Category(val)
    elif dtype == 'bool':
        return bool(val)


def dtype_to_pandas_dtype(dtype: str):
    if dtype == "integer":
        dtype = "int"
    elif dtype == "string":
        dtype = "str"
    elif dtype == "boolean":
        dtype = "bool"

    return dtype


def _get_partitions_and_types(first_file_key: str, bucket):
    ''' Fetch a list of all the partitions actually there and their 
    datatypes. List may be different than passed list if not being used
    for filtering on.
    - This is pulled from the metadata. It is assumed that this package
        is being used with its own publish method.
    '''
    parts_and_types = []
    s3_client = boto3.client('s3')

    first_file = s3_client.head_object(
        Bucket=bucket,
        Key=first_file_key
    )

    partition_metadata = ast.literal_eval(
        first_file['Metadata']['partition_data_types'])

    return partition_metadata


def _get_all_files_list(bucket, key) -> list:
    ''' Get a list of all files to get all partitions values.
    Necesarry to catch all partition values for non-filtered partiions.
    '''
    objects_in_bucket = []
    s3_client = boto3.client('s3')
    paginator = s3_client.get_paginator('list_objects')
    operation_parameters = {'Bucket': bucket,
                            'Prefix': key}
    page_iterator = paginator.paginate(**operation_parameters)
    for page in page_iterator:
        for item in page['Contents']:
            if item['Key'].endswith('.parquet'):
                objects_in_bucket.append(item['Key'])

    return objects_in_bucket


def _parse_partitions_and_values(file_paths: List[str], key: str) -> dict:
    ''' Take the list of all the file keys and return a dict of the
    partitions with arrays of their values.
    '''
    # TODO: find more neat/efficient way to do this
    parts = OrderedDict()
    key_len = len(key)
    for file_path in file_paths:
        # Delete key, split the parts out, delete the file name
        file_path = file_path[key_len:]
        unparsed_parts = file_path.split("/")
        del unparsed_parts[-1]

        for part in unparsed_parts:
            if "=" in part:
                key, value = part.split("=")
                if key not in parts:
                    parts.update({key: set([value])})
                else:
                    parts[key].add(value)

    return parts


def _get_partition_value_data_types(parsed_parts: dict, part_types: dict):
    ''' Convert the collected values to their python data types for use.
    '''
    for part, values in parsed_parts.items():
        part_type = part_types[part]
        if (part_type == 'string') or (part_type == 'category'):
            continue
        elif (part_type == 'int') or (part_type == 'integer'):
            parsed_parts[part] = set(map(int, values))
        elif part_type == 'float':
            parsed_parts[part] = set(map(float, values))
        elif part_type == 'datetime':
            parsed_parts[part] = set(
                map(lambda s: datetime.datetime.strptime(s, "%Y-%m-%d %H:%M:%S"), values))
        elif (part_type == 'bool') or (part_type == 'boolean'):
            parsed_parts[part] = set(map(bool, values))

    return parsed_parts

# TODO: Neaten up?


def _get_filtered_key_list(typed_parts: dict, filters, key) -> List[str]:
    ''' Create list of all "paths" to files after the filtered partitions
    are set ie all non-matching partitions are excluded.
    '''
    filter_keys = []

    for part, part_values in typed_parts.items():
        for f in filters:
            if (f['partition'] == part):
                comparison = OPS[f['comparison']]
                for v in f['values']:
                    typed_parts[part] = set(filter(
                        lambda x: comparison(x, v),
                        typed_parts[part]
                    )
                    )

    def construct_paths(typed_parts, previous_fil_keys: List[str]) -> None:
        if len(typed_parts) > 0:
            part = typed_parts.popitem(last=False)
            new_filter_keys = list()
            for value in part[1]:
                mapped_keys = list(map(
                    (lambda x: str(x) +
                     str(part[0]) + "=" + str(value) + "/"),
                    previous_fil_keys
                ))
                new_filter_keys = new_filter_keys + mapped_keys

            construct_paths(typed_parts, new_filter_keys)
        else:
            filter_keys.append(previous_fil_keys)

    construct_paths(typed_parts, [f"{key}/"])

    # TODO: fix the below mess with random array
    return filter_keys[0]


def _get_filtered_data(bucket: str, paths: list, partition_metadata, parallel=True) -> pd.DataFrame:
    ''' Pull all filtered parquets down and return a dataframe.
    '''

    temp_frames = []

    def append_to_temp(frame):
        temp_frames.append(frame)

    for path in paths:
        if parallel:
            with get_context("spawn").Pool() as pool:
                for path in paths:
                    append_to_temp(
                        pool.apply_async(_s3_parquet_to_dataframe, args=(bucket, path, partition_metadata)).get())
                pool.close()
                pool.join()
        else:
            append_to_temp(_s3_parquet_to_dataframe(
                bucket, path, partition_metadata))

    return pd.concat(temp_frames)


def _s3_parquet_to_dataframe(bucket: str, key: str, partition_metadata) -> pd.DataFrame:
    """ grab a parquet file from s3 and convert to pandas df, add it to the destination"""
    s3 = s3fs.S3FileSystem()
    uri = f"{bucket}/{key}"
    table = pq.ParquetDataset(uri, filesystem=s3)
    frame = table.read_pandas().to_pandas()
    partitions = _repopulate_partitions(uri, partition_metadata)
    for k, v in partitions.items():
        frame[k] = v
    return frame


def _repopulate_partitions(partition_string: str, partition_metadata) -> tuple:
    """ for each val in the partition string creates a list that can be added back into the dataframe"""
    raw = partition_string.split('/')
    partitions = {}
    for string in raw:
        if '=' in string:
            k, v = string.split('=')
            partitions[k] = v

    for key, val in partitions.items():
        partitions[key] = convert_type(val, partition_metadata[key])
    return partitions


def _validate_filter_rules(filters: List[type(Filter)]) -> None:
    ''' Validate that the filters are the correct format and follow basic
    comparison rules.
    '''

    single_value_comparisons = [
        ">",
        "<",
        "<=",
        ">="
    ]
    for f in filters:
        if not all(key in f for key in ("partition", "comparison", "values")):
            raise ValueError(
                "Filters require partition, comparison, and values.")
        elif f["comparison"] not in OPS:
            raise ValueError(
                f"Comparison {f['comparison']} is not supported.")
        elif (f["comparison"] in single_value_comparisons) & (len(f["values"]) != 1):
            raise ValueError(
                f"Comparison {f['comparison']} can only be used with one filter value.")


def _validate_matching_filter_data_type(part_types, filters) -> None:
    ''' Validate that the filters passed are matching to the partitions'
    listed datatypes. This includes validating comparisons too.
    '''
    num_comparisons = [
        ">",
        "<",
        "<=",
        ">="
    ]

    for f in filters:
        try:
            fil_part = part_types[f["partition"]]
        except:
            raise ValueError("Filter does not have a matching partition.")

        if (f["comparison"] in num_comparisons):
            if fil_part in NON_NUM_TYPES:
                raise ValueError(
                    f"Comparison {f['comparison']} cannot be used on partition types of {fil_part}")



import boto3
from s3parq.publish_parq import publish
from s3parq.fetch_parq import fetch, get_max_partition_value, fetch_diff

import pandas as pd
import sys
import logging
from typing import Iterable


class S3Parq:

    def publish(self,
                bucket: str,
                key: str,
                dataframe: pd.DataFrame,
                partitions: Iterable[str]) -> None:

        pub = publish(
            dataframe=dataframe,
            bucket=bucket,
            key=key,
            partitions=partitions
        )
        pub.publish()

    def fetch(self,
              bucket: str,
              key: str,
              **kwargs
              ) -> None:

        return fetch(key=key,
                     bucket=bucket,
                     filters=kwargs.get('partitions', dict())
                     )

    def fetch_diff(self,
        input_bucket: str, 
        input_key: str, 
        comparison_bucket: str, 
        comparison_key: str, 
        partition: str, 
        parallel: bool = True
    ) -> pd.DataFrame:
        return fetch_diff(
            input_bucket = input_bucket, 
            input_key = input_key, 
            comparison_bucket = comparison_bucket, 
            comparison_key = comparison_key, 
            partition = partition, 
            parallel = parallel
        )

    def get_max_partition_value(self, bucket: str, key: str, partition: str) -> any:
        return get_max_partition_value(
            bucket=bucket,
            key=key,
            partition=partition
        )



from setuptools import setup, find_packages
from os import path

package_name = "s3parq"
package_version = "0.0.3"
description = "Write and read/query s3 parquet data using Athena/Spectrum/Hive style partitioning."
cur_directory = path.abspath(path.dirname(__file__))
with open(path.join(cur_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Integrichain Innovation Team",
    author_email="engineering@integrichain.com",
    url="https://github.com/IntegriChain1/s3parq",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7"
        ],
    packages= find_packages(exclude=("tests",)),
    include_package_data=True,
    tests_require = ["dfmock==0.0.14","moto"],
    install_requires=["pandas","pyarrow","boto3","s3fs"]
    )
